#ifndef PYTHONIC_INCLUDE_NUMPY_BITWISE_OR_REDUCE_HPP
#define PYTHONIC_INCLUDE_NUMPY_BITWISE_OR_REDUCE_HPP

#define UFUNC_NAME bitwise_or
#define UFUNC_INAME ior
#include "pythonic/include/numpy/ufunc_reduce.hpp"
#undef UFUNC_NAME
#undef UFUNC_INAME

#endif
