#ifndef PYTHONIC_OPERATOR_IXOR_HPP
#define PYTHONIC_OPERATOR_IXOR_HPP

#include "pythonic/include/operator_/ixor.hpp"

#define OPERATOR_NAME ixor
#define OPERATOR_SYMBOL ^
#define OPERATOR_ISYMBOL ^=

#include "pythonic/operator_/icommon.hpp"

#endif
