#ifndef PYTHONIC_INCLUDE_NUMPY_BITWISE_OR_ACCUMULATE_HPP
#define PYTHONIC_INCLUDE_NUMPY_BITWISE_OR_ACCUMULATE_HPP

#define UFUNC_NAME bitwise_or
#include "pythonic/include/numpy/ufunc_accumulate.hpp"

#endif
