#ifndef PYTHONIC_INCLUDE_OPERATOR_IMUL_HPP
#define PYTHONIC_INCLUDE_OPERATOR_IMUL_HPP

#define OPERATOR_NAME imul
#define OPERATOR_SYMBOL *
#define OPERATOR_ISYMBOL *=
#include "pythonic/include/operator_/icommon.hpp"

#endif
