#ifndef PYTHONIC_IO__IO_TEXTIOWRAPPER_READLINES_HPP
#define PYTHONIC_IO__IO_TEXTIOWRAPPER_READLINES_HPP

#include "pythonic/include/io/_io/TextIOWrapper/readlines.hpp"
#include "pythonic/builtins/file/readlines.hpp"

#endif
