#ifndef PYTHONIC_INCLUDE_BUILTIN_TRUE_HPP
#define PYTHONIC_INCLUDE_BUILTIN_TRUE_HPP

#include "pythonic/include/types/bool.hpp"

PYTHONIC_NS_BEGIN

namespace builtins
{

  static const bool True = true;
}
PYTHONIC_NS_END

#endif
