#ifndef PYTHONIC_INCLUDE_OPERATOR_ILSHIFT_HPP
#define PYTHONIC_INCLUDE_OPERATOR_ILSHIFT_HPP

#define OPERATOR_NAME ilshift
#define OPERATOR_SYMBOL <<
#define OPERATOR_ISYMBOL <<=
#include "pythonic/include/operator_/icommon.hpp"

#endif
